﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Security.Cryptography.X509Certificates;
using System.Net.Sockets;
using System.Security.Cryptography;
using System.Windows.Forms;
using System.IO;

namespace MySSLLib
{
  public enum PackageType : byte
  {
    Cert = 1,
    SessionKey = 2,
    ApplicationData = 3,
    Error = 4
  }

  public class StateObject
  {
    public Socket workSocket = null;
    public const int BUFFER_SIZE = 1024 * 64;
    public byte[] buffer = new byte[BUFFER_SIZE];

    public X509Certificate2 remoteCert;
    public X509Certificate2 localCert;
    public byte[] localSessionKey;
    public byte[] remoteSessionKey;
    public byte[] sentData;
    public byte[] recvData;
  }

  public class SslHelper
  {
    public static PackageType Handle(StateObject so, int read)
    {
      switch ((PackageType)so.buffer[0])
      {
        case PackageType.Cert:
          {
            byte[] data = RemoveType(so.buffer, read);
            so.remoteCert = new X509Certificate2(data);
            Console.WriteLine("Recv<<<<<<<<<<<<<<<<<<<<<<");
            PrintCert(so.remoteCert);
            break;
          }

        case PackageType.SessionKey:
          {
            byte[] data = RemoveType(so.buffer, read);
            RSACryptoServiceProvider rsa = (RSACryptoServiceProvider)so.localCert.PrivateKey;
            so.remoteSessionKey = rsa.Decrypt(data, false);
            Console.WriteLine("Recv<<<<<<<<<<<<<<<<<<<<<<");
            PrintSessionKey(so.remoteSessionKey);
            break;
          }

        case PackageType.ApplicationData:
          {
            byte[] data = RemoveType(so.buffer, read);

            // Decrypt
            DESCryptoServiceProvider des = new DESCryptoServiceProvider();
            ICryptoTransform cTransform = des.CreateDecryptor(so.localSessionKey, new byte[8]);
            int len = BitConverter.ToInt32(data, 0);
            byte[] cipherText = new byte[len];
            Array.Copy(data, 4, cipherText, 0, len);
            MemoryStream ms2 = new MemoryStream(cipherText);
            CryptoStream decStream = new CryptoStream(ms2, cTransform, CryptoStreamMode.Read);

            Byte[] tempBuffer = new Byte[1024];
            Byte[] outputBytes = new byte[0];
            int nRead = decStream.Read(tempBuffer, 0, tempBuffer.Length);
            int nLength = 0;
            while (0 != nRead)
            {
              nLength = outputBytes.Length;
              Array.Resize(ref outputBytes, nLength + nRead);
              Array.Copy(tempBuffer, 0, outputBytes, nLength, nRead);
              nRead = decStream.Read(tempBuffer, 0, tempBuffer.Length);
            }
            so.recvData = outputBytes;
            Console.WriteLine("Recv<<<<<<<<<<<<<<<<<<<<<<");
            PrintData(so.recvData);

            // hash
            MD5CryptoServiceProvider md5 = new MD5CryptoServiceProvider();
            byte[] hash = md5.ComputeHash(so.recvData);

            // verify
            int signlen = BitConverter.ToInt32(data, len + 4);
            byte[] sign = new byte[signlen];
            Array.Copy(data, 4+len+4,sign, 0, signlen);
            RSACryptoServiceProvider rsa = (RSACryptoServiceProvider)so.remoteCert.PublicKey.Key;
            try
            {
              bool verify = rsa.VerifyHash(hash, CryptoConfig.MapNameToOID("MD5"), sign);
              if (verify)
              {
                Console.WriteLine("verify correct!");
              }
              else
              {
                Console.WriteLine("verify error!");
              }
            }
            catch
            {
              Console.WriteLine("verify error!");
              return PackageType.Error;
            }
              break;
          }
      }

      return (PackageType)so.buffer[0];

    }

    public static void Send(StateObject so, PackageType type)
    {
      switch (type)
      {
        case PackageType.Cert:
          {
            byte[] certBytes = so.localCert.Export(X509ContentType.Cert);
            byte[] sentBytes = AddType(certBytes, type);
            so.workSocket.Send(sentBytes, 0, sentBytes.Length, SocketFlags.None);
            Console.WriteLine("Send>>>>>>>>>>>>>>>>>>>>>");
            PrintCert(so.localCert);
            break;
          }

        case PackageType.SessionKey:
          {
            RSACryptoServiceProvider rsa = (RSACryptoServiceProvider)so.remoteCert.PublicKey.Key;
            byte[] encrypedSessionKey = rsa.Encrypt(so.localSessionKey, false);
            byte[] sentBytes = AddType(encrypedSessionKey, type);
            so.workSocket.Send(sentBytes, 0, sentBytes.Length, SocketFlags.None);
            Console.WriteLine("Send>>>>>>>>>>>>>>>>>>>>>");
            PrintSessionKey(so.localSessionKey);
            break;
          }

        case PackageType.ApplicationData:
          {
            MemoryStream memoryStream = new MemoryStream();

            // encrypt
            DESCryptoServiceProvider des = new DESCryptoServiceProvider();
            ICryptoTransform cTransform = des.CreateEncryptor(so.remoteSessionKey, new byte[8]);
            MemoryStream ms = new MemoryStream();
            CryptoStream encStream = new CryptoStream(ms, cTransform, CryptoStreamMode.Write);
            encStream.Write(so.sentData, 0, so.sentData.Length);
            encStream.FlushFinalBlock();

            byte[] temp = ms.ToArray();
            memoryStream.Write(BitConverter.GetBytes(temp.Length), 0, 4);
            memoryStream.Write(temp, 0, temp.Length);

            // hash
            MD5CryptoServiceProvider md5 = new MD5CryptoServiceProvider();
            byte[] hash = md5.ComputeHash(so.sentData);

            // sign
            RSACryptoServiceProvider rsa = (RSACryptoServiceProvider) so.localCert.PrivateKey;
            byte[] sign = rsa.SignHash(hash, CryptoConfig.MapNameToOID("MD5"));

            memoryStream.Write(BitConverter.GetBytes(sign.Length), 0, 4);
            memoryStream.Write(sign, 0, sign.Length);

            byte[] sentBytes = AddType(memoryStream.ToArray(), type);
            so.workSocket.Send(sentBytes, 0, sentBytes.Length, SocketFlags.None);
            Console.WriteLine("Send>>>>>>>>>>>>>>>>>>>>>");
            PrintData(so.sentData);
            break;
          }
      }

    }

    private static byte[] RemoveType(byte[] input, int len)
    {
      byte[] output = new byte[len - 1];
      Array.Copy(input, 1, output, 0, output.Length);
      return output;
    }

    private static byte[] AddType(byte[] input, PackageType type)
    {
      byte[] output = new byte[input.Length + 1];
      output[0] = (byte)type;
      Array.Copy(input, 0, output, 1, input.Length);
      return output;
    }

    public static X509Certificate2 GetCert()
    {
      Console.WriteLine("Press S to get cert from store");
      Console.WriteLine("Press F to read cert from file");
      ConsoleKeyInfo keyInfo = Console.ReadKey(true);
      switch (keyInfo.KeyChar)
      {
        case 'f':
        case 'F':
          OpenFileDialog dialog = new OpenFileDialog();
          dialog.Filter = "Cert files(*.pfx;*.p7b;*.cer)|*.pfx;*.p7b;*.cer|All files(*.*)|*.*";
          DialogResult dr = dialog.ShowDialog();
          if (dr == DialogResult.OK)
          {
            Console.WriteLine("Input the password for the cert:");
            StringBuilder stringBuilder = new StringBuilder();
            while (true)
            {
              ConsoleKeyInfo passInfo = Console.ReadKey(true);
              if (passInfo.Key == ConsoleKey.Enter)
              {
                break;
              }
              stringBuilder.Append(passInfo.KeyChar);
            }
            return new X509Certificate2(dialog.FileName, stringBuilder.ToString(), X509KeyStorageFlags.Exportable);
          }
          else
          {
            return null;
          }
        case 'S':
        case 's':
          X509Store store = new X509Store(StoreName.My, StoreLocation.LocalMachine);
          store.Open(OpenFlags.ReadOnly);
          X509Certificate2Collection collection = X509Certificate2UI.SelectFromCollection(store.Certificates,
                                                                                          "Select Certificate",
                                                                                          "Please select a certificate from the following list",
                                                                                          X509SelectionFlag.SingleSelection);
          if (collection.Count == 0)
          {
            return null;
          }
          else
          {
            return collection[0];
          }
        default:
          return null;
      }
    }

    public static string ConvertBytesToHexString(byte[] bytes)
    {
      if (bytes == null || bytes.Length == 0)
      {
        return string.Empty;
      }
      StringBuilder sb = new StringBuilder();
      foreach (byte temp in bytes)
      {
        sb.Append(string.Format("{0:X2} ", temp));
      }
      sb.Remove(sb.Length - 1, 1);
      sb.Append(string.Format(("[{0}]"), bytes.Length));

      return sb.ToString();
    }

    public static void PrintSessionKey(byte[] key)
    {
      Console.WriteLine("============================================");
      Console.WriteLine("session key");
      Console.WriteLine("============================================");
      Console.WriteLine("Session Key: {1}", Environment.NewLine, ConvertBytesToHexString(key));
    }

    private static void PrintCert(X509Certificate2 x509)
    {
      Console.WriteLine("============================================");
      Console.WriteLine("cert");
      Console.WriteLine("============================================");

      //Print to console information contained in the certificate.
      Console.WriteLine("Subject: {1}", Environment.NewLine, x509.Subject);
      Console.WriteLine("Issuer: {1}", Environment.NewLine, x509.Issuer);
      Console.WriteLine("Version: {1}", Environment.NewLine, x509.Version);
      Console.WriteLine("Valid Date: {1}", Environment.NewLine, x509.NotBefore);
      Console.WriteLine("Expiry Date: {1}", Environment.NewLine, x509.NotAfter);
      Console.WriteLine("Thumbprint: {1}", Environment.NewLine, x509.Thumbprint);
      Console.WriteLine("Serial Number: {1}", Environment.NewLine, x509.SerialNumber);
      Console.WriteLine("Friendly Name: {1}", Environment.NewLine, x509.PublicKey.Oid.FriendlyName);
      Console.WriteLine("Public Key Format: {1}", Environment.NewLine, x509.PublicKey.EncodedKeyValue.Format(true));
      //Console.WriteLine("Raw Data Length: {1}", Environment.NewLine, x509.RawData.Length);
      //Console.WriteLine("Certificate to string: {1}", Environment.NewLine, x509.ToString(true));

      //Console.WriteLine("Certificate to XML String: {1}", Environment.NewLine, x509.PublicKey.Key.ToXmlString(false));

    }

    private static void PrintData(byte[] data)
    {
      Console.WriteLine("============================================");
      Console.WriteLine("data");
      Console.WriteLine("============================================");

      Console.WriteLine(Encoding.ASCII.GetString(data));
    }
  }
}
