﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Security.Cryptography.X509Certificates;
using System.Net.Sockets;
using System.Security.Cryptography;
using System.Net;
using System.Threading;

namespace MySSLLib
{
  public class SslSocket
  {
    private Socket socket = null;
    private int port = 443;
    public X509Certificate2 localCertificate;
    private StateObject clientSo;
    private Dictionary<string, StateObject> serverSoDict;

    #region server

    public void StartAsServer(X509Certificate2 cert)
    {
      localCertificate = cert;
      serverSoDict = new Dictionary<string, StateObject>();
      socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
      socket.Bind(new IPEndPoint(IPAddress.Any, port));
      socket.Listen(5);
      socket.BeginAccept(new AsyncCallback(OnClientConnected), null);
    }

    private void OnClientConnected(IAsyncResult ar)
    {
      Socket clientSocket = socket.EndAccept(ar);
      socket.BeginAccept(new AsyncCallback(OnClientConnected), null);

      StateObject so = new StateObject();
      string remoteAddress = clientSocket.RemoteEndPoint.ToString();
      so.workSocket = clientSocket;
      so.localCert = localCertificate;
      DESCryptoServiceProvider des = new DESCryptoServiceProvider();
      so.localSessionKey = des.Key;
      //SslHelper.PrintSessionKey(des.Key);
      serverSoDict.Add(remoteAddress, so);

      clientSocket.BeginReceive(so.buffer, 0, StateObject.BUFFER_SIZE, SocketFlags.None, new AsyncCallback(OnDataReceivedServer), so);
    }

    private void OnDataReceivedServer(IAsyncResult ar)
    {
      StateObject so = (StateObject)ar.AsyncState;
      Socket clientSocket = so.workSocket;
      int read = clientSocket.EndReceive(ar);
      if (read > 0)
      {
        PackageType type = SslHelper.Handle(so, read);
        if (type == PackageType.Cert)
        {
          SslHelper.Send(so, PackageType.Cert);
          Thread.Sleep(500);
          SslHelper.Send(so, PackageType.SessionKey);
        }
        else if (type == PackageType.SessionKey)
        {
          //SslHelper.Send(so, 0x03);
        }
        else if (type == PackageType.ApplicationData)
        {
          string message = "Response to client.";
          so.sentData = Encoding.ASCII.GetBytes(message);
          SslHelper.Send(so, PackageType.ApplicationData);
        }
      }
      clientSocket.BeginReceive(so.buffer, 0, StateObject.BUFFER_SIZE, SocketFlags.None, new AsyncCallback(OnDataReceivedServer), so);
      

    }

    //public void Send(byte[] data, string address)
    //{
    //  StateObject so = serverSoDict[address];
    //  if (so != null)
    //  {
    //    so.sentData = data;
    //    SslHelper.Send(so, PackageType.ApplicationData);
    //  }
    //}

    #endregion

    #region client
    
    public void StartAsClient(X509Certificate2 cert, string ip)
    {
      localCertificate = cert;
      socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);

      socket.Connect(IPAddress.Parse(ip), port);

      clientSo = new StateObject();
      clientSo.workSocket = socket;
      clientSo.localCert = cert;
      DESCryptoServiceProvider des = new DESCryptoServiceProvider();
      clientSo.localSessionKey = des.Key;
      //SslHelper.PrintSessionKey(des.Key);

      SslHelper.Send(clientSo, PackageType.Cert);
      socket.BeginReceive(clientSo.buffer, 0, StateObject.BUFFER_SIZE, SocketFlags.None, new AsyncCallback(OnDataReceivedClient), clientSo);
    }

    private void OnDataReceivedClient(IAsyncResult ar)
    {
      StateObject so = (StateObject)ar.AsyncState;
      Socket clientSocket = so.workSocket;
      int read = clientSocket.EndReceive(ar);
      if (read > 0)
      {
        PackageType type = SslHelper.Handle(so, read);
        if (type == PackageType.Cert)
        {
          SslHelper.Send(so, PackageType.SessionKey);
        }
        else if (type == PackageType.SessionKey)
        {
          //SslHelper.Send(so, 0x03);
        }
        else if (type == PackageType.ApplicationData)
        {

        }
      }
      clientSocket.BeginReceive(so.buffer, 0, StateObject.BUFFER_SIZE, SocketFlags.None, new AsyncCallback(OnDataReceivedClient), so);

    }

    public void Send(byte[] data)
    {
      if (clientSo != null)
      {
        // client
        clientSo.sentData = data;
        SslHelper.Send(clientSo, PackageType.ApplicationData);
      }
    }

    public void Close()
    {
      socket.Close();
    }

    #endregion
  }
}
